package com.sencha.gxt.cell.core.client.form;

import java.text.ParseException;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.google.gwt.cell.client.ValueUpdater;
import com.google.gwt.core.client.GWT;
import com.google.gwt.core.client.Scheduler;
import com.google.gwt.core.client.Scheduler.ScheduledCommand;
import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.InputElement;
import com.google.gwt.dom.client.NativeEvent;
import com.google.gwt.event.dom.client.KeyCodes;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.safehtml.shared.SafeHtmlBuilder;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.Event.NativePreviewEvent;
import com.google.gwt.user.client.Event.NativePreviewHandler;
import com.sencha.gxt.core.client.dom.XElement;
import com.sencha.gxt.widget.core.client.event.ParseErrorEvent;
import com.sencha.gxt.widget.core.client.event.TriggerClickEvent;

public class TriggerFieldCell<T> extends ValueBaseFieldCell<T> {

  public interface TriggerFieldAppearance extends ValueBaseFieldAppearance {
    void onTriggerClick(XElement parent, boolean click);

    void onTriggerOver(XElement parent, boolean over);

    void render(SafeHtmlBuilder sb, String value, FieldAppearanceOptions options);

    void setEditable(XElement parent, boolean editable);

    void setTriggerVisible(XElement parent, boolean visible);

    boolean triggerIsOrHasChild(XElement parent, Element target);
  }

  private class MouseDownHandler implements NativePreviewHandler {

    private Context context;
    private final Element parent;
    private ValueUpdater<T> updater;
    private HandlerRegistration reg;

    public MouseDownHandler(Context context, ValueUpdater<T> updater, Element parent) {
      logger.log(Level.FINEST, "TriggerFieldCell add event preview");
      this.context = context;
      this.updater = updater;
      this.parent = parent;
      this.reg = Event.addNativePreviewHandler(this);

      focusManagerRegistration = reg;
    }

    public void onPreviewNativeEvent(NativePreviewEvent event) {
      NativeEvent e = event.getNativeEvent();
      XElement target = event.getNativeEvent().getEventTarget().cast();
      if ("mousedown".equals(e.getType())) {
        logger.log(Level.FINEST, "TriggerFieldCell preview mouse down");
        if (!isFocusClick(parent.<XElement> cast(), target)) {
          logger.log(Level.FINEST, "preview mouse down not a focus click, remove preview");
          reg.removeHandler();
          focusManagerRegistration = null;
          lastValueUpdater = updater;
          lastContext = context;
          triggerBlur(context, parent.<XElement> cast(), lastValue, updater);
        }
      }
    }
  }

  protected final TriggerFieldAppearance appearance;
  protected boolean mimicing;

  private boolean editable = true;
  private boolean hideTrigger;
  private boolean monitorTab = true;
  private HandlerRegistration focusManagerRegistration;
  private Logger logger = Logger.getLogger(getClass().getName());

  public TriggerFieldCell() {
    this(GWT.<TriggerFieldAppearance> create(TriggerFieldAppearance.class));
  }

  public TriggerFieldCell(TriggerFieldAppearance appearance) {
    super(appearance);
    this.appearance = appearance;
  }

  /**
   * Returns true if the field is editable.
   * 
   * @return true if editable
   */
  public boolean isEditable() {
    return editable;
  }

  /**
   * Returns true if the trigger is hidden.
   * 
   * @return true if hidden
   */
  public boolean isHideTrigger() {
    return hideTrigger;
  }

  /**
   * Returns true if tab key events are being monitored.
   * 
   * @return true if monitoring
   */
  public boolean isMonitorTab() {
    return monitorTab;
  }

  public void onBrowserEvent(Context context, Element parent, T value, NativeEvent event, ValueUpdater<T> valueUpdater) {
    super.onBrowserEvent(context, parent, value, event, valueUpdater);
    String eventType = event.getType();
    Object key = context.getKey();
    if ("click".equals(eventType)) {
      onClick(context, parent.<XElement> cast(), event, value, valueUpdater);
    } else if ("keyup".equals(eventType)) {
      // Record keys as they are typed.
      FieldViewData vd = getViewData(key);
      if (vd == null) {
        vd = new FieldViewData(getText(XElement.as(parent)));
        setViewData(key, vd);
      }
      vd.setCurrentValue(getText(XElement.as(parent)));
    }
  }

  @Override
  public void render(Context context, T value, SafeHtmlBuilder sb) {
    String v = "";
    if (value != null) {
      v = getPropertyEditor().render(value);
    }

    FieldViewData viewData = checkViewData(context, v);

    FieldAppearanceOptions options = new FieldAppearanceOptions(width, height, isReadOnly());
    options.setEmptyText(getEmptyText());
    options.setHideTrigger(isHideTrigger());
    options.setEditable(editable);

    String s = (viewData != null) ? viewData.getCurrentValue() : v;
    appearance.render(sb, s == null ? "" : s, options);

  }

  public void setEditable(XElement parent, boolean editable) {
    this.editable = editable;

    appearance.setEditable(parent, editable);

    InputElement fromEl = appearance.getInputElement(parent).cast();
    if (!isReadOnly()) {
      fromEl.setPropertyBoolean("readOnly", !editable);
    }
  }

  public void setHideTrigger(boolean hideTrigger) {
    this.hideTrigger = hideTrigger;
  }

  /**
   * True to monitor tab key events to force the bluring of the field (defaults
   * to true).
   * 
   * @param monitorTab true to monitor tab key events
   */
  public void setMonitorTab(boolean monitorTab) {
    this.monitorTab = monitorTab;
  }

  @Override
  public void finishEditing(Element parent, T value, Object key, ValueUpdater<T> valueUpdater) {
    logger.log(Level.FINEST, "finishEditing");

    String newValue = getText(XElement.as(parent));

    // Get the view data.
    FieldViewData vd = getViewData(key);
    if (vd == null) {
      vd = new FieldViewData(newValue);
      setViewData(key, vd);
    }
    vd.setCurrentValue(newValue);

    // Fire the value updater if the value has changed.
    if (valueUpdater != null && !vd.getCurrentValue().equals(vd.getLastValue())) {
      vd.setLastValue(newValue);
      try {
        logger.log(Level.FINEST, "finishEditing saving value");
        valueUpdater.update(newValue.equals("") ? null : getPropertyEditor().parse(newValue));
      } catch (ParseException e) {
        logger.log(Level.FINEST, "finishEditing parseError: " + e.getMessage());
        vd.setCurrentValue("");
        valueUpdater.update(null);
        setText(parent.<XElement> cast(), "");
        fireEvent(new ParseErrorEvent(newValue, e));
      }
    } else {
      logger.log(Level.FINEST, "finishEditing value not changed: " + newValue + " old: " + vd.getLastValue());
    }

    // Blur the element.
    // calling super.finishEditing causes the input element to be blurred which is stopping focus of the the field
    // that was clicked
    clearViewData(key);
    clearFocusKey();
    super.onBlur(lastContext, parent.<XElement> cast(), value, null);
  }
  
  private native void clearFocusKey() /*-{
    this.@com.google.gwt.cell.client.AbstractInputCell::focusedKey = null;
  }-*/;


  protected boolean isFocusClick(XElement parent, XElement target) {
    return parent.isOrHasChild(target);
  }

  @Override
  protected void onBlur(Context context, XElement parent, T value, NativeEvent event) {
    // do nothing
    logger.log(Level.FINEST, "TriggerFieldCell onBlur - mimiciing = " + mimicing);

    if (!mimicing) {
      logger.log(Level.FINEST, "TriggerFieldCell onBlur no mimicing so calling super");
      super.onBlur(context, parent, value, event);
    }
  }

  protected void onClick(Context context, XElement parent, NativeEvent event, T value, ValueUpdater<T> updater) {
    Element target = event.getEventTarget().cast();

    if (!isReadOnly() && !isReadOnly() && appearance.triggerIsOrHasChild(parent.<XElement> cast(), target)) {
      onTriggerClick(context, parent, event, value, updater);
    }

    if (!editable) {
      event.preventDefault();
      event.stopPropagation();
    }
  }

  @Override
  protected void onFocus(Context context, XElement parent, T value, NativeEvent event, ValueUpdater<T> valueUpdater) {
    logger.log(Level.FINEST, "TriggerFieldCell onFocus " + parent.getId());
    if (mimicing) {
      logger.log(Level.FINEST, "TriggerFieldCell onFocus mimic = true so exiting " + parent.getId());
      return;
    }

    if (!mimicing) {
      super.onFocus(context, parent, value, event, valueUpdater);
      logger.log(Level.FINEST, "onFocus setting mimicing to true " + parent.getId());
      saveContext(context, parent, event, valueUpdater, value);
      mimicing = true;
      appearance.onFocus(parent, true);
      new MouseDownHandler(context, valueUpdater, parent);
    }
  }

  protected void onKeyDown(final Context context, final Element parent, final T value, NativeEvent event,
      final ValueUpdater<T> valueUpdater) {
    super.onKeyDown(context, parent, value, event, valueUpdater);
    if (monitorTab && event.getKeyCode() == KeyCodes.KEY_TAB) {
      logger.log(Level.FINEST, "onKeyDown Tab " + parent.getId());
      Scheduler.get().scheduleDeferred(new ScheduledCommand() {
        @Override
        public void execute() {
          if (hasFocus(context, XElement.as(parent))) {
            triggerBlur(context, parent.<XElement> cast(), value, valueUpdater);
          }
        }
      });
    }
  }

  @Override
  protected void onMouseOut(XElement parent, NativeEvent event) {
    super.onMouseOut(parent, event);
    XElement target = event.getEventTarget().cast();
    if (appearance.triggerIsOrHasChild(parent.<XElement> cast(), target)) {
      appearance.onTriggerOver(parent.<XElement> cast(), false);
    }
  }

  @Override
  protected void onMouseOver(XElement parent, NativeEvent event) {
    super.onMouseOver(parent, event);
    XElement target = event.getEventTarget().cast();
    if (appearance.triggerIsOrHasChild(parent.<XElement> cast(), target)) {
      appearance.onTriggerOver(parent.<XElement> cast(), true);
    }
  };

  protected void onTriggerClick(Context context, XElement parent, NativeEvent event, T value, ValueUpdater<T> updater) {
    fireEvent(context, new TriggerClickEvent());
  }

  protected void triggerBlur(Context context, final XElement parent, T value, ValueUpdater<T> valueUpdater) {
    logger.log(Level.FINEST, "TriggerFieldCell triggerBlur " + parent.getId());
    logger.log(Level.FINEST, "TriggerFieldCell triggerBlur mimicing = false");

    if (focusManagerRegistration != null) {
      logger.log(Level.FINEST, "TriggerFieldCell triggerBlur remove event preview");
      focusManagerRegistration.removeHandler();
      focusManagerRegistration = null;
    }

    mimicing = false;
    finishEditing(parent, value, context.getKey(), valueUpdater);

  }

  protected boolean validateBlur(Element target) {
    return true;
  }

}
